import re
import subprocess
import sys
from pathlib import Path
from typing import Optional

BASE_METADATA = {
    "nameInLink": "true",
    "link-citations": "true",
    "linkReferences": "true",
    "reference-section-title": "References",
    "figPrefix": "Figure",
    "eqnPrefix": "Equation",
    "tblPrefix": "Table",
    "lstPrefix": "List",
    "secPrefix": "Section",
}


def run_pandoc(
    input_path: Path,
    template_path: Path,
    bib_path: Optional[Path] = None,
    metadata_path: Optional[Path] = None,
    default_img_ext: Optional[str] = None,
    natbib: bool = False,
    verbose: bool = False,
    **metadata,
) -> str:
    """Run `pandoc` to convert input markdown to LaTeX.

    Args:
        input_path: Path to input file.
        template_path: Path to template file.
        bib_path: Path to bibliography file.
        metadata_path: Path to metadata file.
        default_img_ext: Default image extension.
        natbib: Whether to use natbib for bibliography.
        verbose: Whether to enable verbose output.
        metadata: Pandoc metadata (each key-value pair will be added using `-M`).

    Returns:
        Generated LaTeX text.

    Raises:
        subprocess.CalledProcessError: If `pandoc` subprocess fails.
    """
    cmd = ["pandoc"]
    cmd.append("--from=markdown")
    cmd.append("--to=latex")
    cmd.append("--filter=pandoc-crossref")
    cmd.append("--citeproc")

    cmd.append("--wrap=none")

    cmd.append(f"--template={template_path}")
    if metadata_path is not None:
        cmd.append(f"--metadata-file={metadata_path}")
    if bib_path is not None:
        cmd.append(f"--bibliography={bib_path}")
        if natbib:
            cmd.append("--natbib")

    if default_img_ext is not None:
        cmd.append(f"--default-image-extension={default_img_ext}")
    if verbose:
        cmd.append("--verbose")

    base_metadata = BASE_METADATA.copy()
    base_metadata.update(metadata)
    for mk, mv in base_metadata.items():
        cmd.append(f"--metadata={mk}={mv}")

    cmd.append(str(input_path))

    if verbose:
        print("+", *cmd, file=sys.stderr)
    else:
        print(f"+ pandoc '{input_path}' -t '{template_path}'", file=sys.stderr)
    pdone = subprocess.run(
        cmd, check=True, text=True, stdout=subprocess.PIPE, stderr=sys.stderr
    )
    output = _fix_tables(_fix_subfigures(pdone.stdout))
    return output


def _fix_tables(latexs):
    """Modify LaTeX text to use regular `table` instead of `longtable`."""

    def _fix(match):
        match_str = match.group(1)

        # Capture caption to move to the end.
        cap_match = re.search(
            r"\\caption(\[.*\])?\{(.*)\}(\\tabularnewline)?", match_str
        )
        cap = ""
        if cap_match:
            cap_match_str, cap = cap_match.group(0), cap_match.group(2)
            match_str = match_str.replace(cap_match_str, "")
            if cap:
                cap = "\\caption{" + cap + "}"

        # Remove everything between '\endfirsthead' and '\endlastfoot'.
        match_str = re.sub(
            r"(\\endfirsthead.*?)?\\endlastfoot", "", match_str, flags=re.DOTALL
        )

        # Replace '\<top/mid/bottom>rule()' with '\<top/mid/bottom>rule'.
        for _rule in ["top", "mid", "bottom"]:
            match_str = re.sub(rf"\\{_rule}rule\(\)", rf"\\{_rule}rule", match_str)

        return (
            "\\begin{table}\n"
            "\\centering\n"
            "\\begin{tabular}"
            f"{match_str.strip()}\n"
            "\\bottomrule\\noalign{}\n"
            "\\end{tabular}\n"
            f"{cap}\n"
            "\\end{table}"
        )

    return re.sub(
        r"\\begin{longtable}(.*?)\\end{longtable}", _fix, latexs, flags=re.DOTALL
    )


def _fix_subfigures(latexs):
    """Modify figures by wrapping sub-figures in an 'adjustbox'."""

    def _fix(match):
        match_str = match.group(1)

        # First extract the subfloats. Each subfloat is on a single line.
        subfloats = re.findall(r"\\subfloat.*\n", match_str)
        # Remove the subfloats from the match string.
        match_str = re.sub(r"\\subfloat.*\n", "", match_str)
        # Clean up the match string.
        match_str = match_str.strip().replace("\n\n", "\n")

        # Add '\quad' between the subfloats.
        subfloats = "\\quad\n".join(subfloats)
        # Wrap the subfloats and the rest of the figure in a 'figure' environment.
        return "\\begin{figure}\n" + subfloats + match_str + "\n\\end{figure}"

    _env = "pandoccrossrefsubfigures"
    return re.sub(
        rf"\\begin{{{_env}}}(.*?)\\end{{{_env}}}", _fix, latexs, flags=re.DOTALL
    )



import importlib.resources
import sys
from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import AbstractContextManager, contextmanager, ExitStack
from importlib.abc import Traversable
from pathlib import Path
from typing import final


class Template(ABC):
    """Base class for LaTeX templates."""

    @abstractmethod
    def load(
        self,
    ) -> AbstractContextManager[tuple[Path, dict[str, Path], dict[str, Path]]]:
        """Load the template, and any associated resources.

        Yields:
            A tuple of (template path, a dictionary mapping static resource names
            to their paths, a dictionary mapping dynamic resource names to paths).
        """

    @staticmethod
    @final
    def make(arg: str) -> "Template":
        """Get a template from a name (`BuiltinTemplate`), or path (`CustomTemplate`).

        Args:
            arg: Name of a builtin template, or path to a custom template.
        """
        if arg in BuiltinTemplate.BUILTIN_TEMPLATE_NAMES:
            return BuiltinTemplate(arg)
        return CustomTemplate(Path(arg))


class BuiltinTemplate(Template):
    _builtin_template_paths: dict[str, Traversable] = {}

    _static_resources = defaultdict(
        list,
        {
            # These are per-template static resources. The '*' key if for resources
            # that are common to all templates.
            "*": ["reqsetup.tex", "floatsetup.tex", "mathsetup.tex", "miscsetup.tex"],
            "basic": ["bibsupersetup.tex"],
            "stylish": ["bibsupersetup.tex"],
            "standalone": ["bibsupersetup.tex"],
            "spacious": ["bibsupersetup.tex"],
            "iccv": [
                "iccv/iccv.sty",
                "iccv/esopic.sty",
                "iccv/everyshi.sty",
                "iccv/bibstyle.bst",
                "bibnonatsetup.tex",
            ],
            "iclr": [
                "iclr/iclr.sty",
                "iclr/bibstyle.bst",
                "iclr/mathcommands.tex",
                "bibnosupersetup.tex",
            ],
            "icml": ["icml/icml.sty", "icml/bibstyle.bst", "bibnosupersetup.tex"],
            "neurips": ["neurips/neurips.sty", "bibnosupersetup.tex"],
        },
    )
    _dynamic_resources = defaultdict(
        list,
        {
            # Dynamic resources, defined similarly to static resources.
            "*": ["addappendices.tex"],
            "basic": ["authsetup.tex"],
            "iccv": ["authsetup.tex", "dblfloatsetup.tex"],
            "iclr": ["authsetup.tex"],
            "icml": ["dblfloatsetup.tex"],
            "neurips": ["authsetup.tex"],
            "spacious": ["authsetupshort.tex"],
            "stylish": ["authsetupshort.tex", "dblfloatsetup.tex"],
        },
    )

    # Get available builtin templates.
    for _template_path in (
        _path
        for _path in (importlib.resources.files(__package__) / "templates").iterdir()
        if _path.name.endswith(".tex")
    ):
        _name = _template_path.name.removesuffix(".tex")
        _builtin_template_paths[_name] = _template_path

    BUILTIN_TEMPLATE_NAMES = list(_builtin_template_paths.keys())

    def __init__(self, name: str):
        if name not in self.BUILTIN_TEMPLATE_NAMES:
            raise ValueError(f"no builtin template named {name!r}")
        self.name = name

    @contextmanager
    def load(self):
        ctx_stk = ExitStack()
        template_path = ctx_stk.enter_context(
            importlib.resources.as_file(self._builtin_template_paths[self.name])
        )

        static_resc_paths = {}
        dynamic_resc_paths = {}

        # Load resources for the template.
        _resources_root = importlib.resources.files(__package__) / "resources"
        for _kind, _kind_src_dict, _kind_target_dict in zip(
            ["static", "dynamic"],
            [self._static_resources, self._dynamic_resources],
            [static_resc_paths, dynamic_resc_paths],
        ):
            for _resc in _kind_src_dict["*"] + _kind_src_dict[self.name]:
                _resc_path = _resources_root / _kind / _resc
                _resc_name = str(_resc_path.with_suffix("").name)
                print(f"+ load resource {_resc_name!r}", file=sys.stderr)
                _kind_target_dict[_resc_name] = ctx_stk.enter_context(
                    importlib.resources.as_file(_resc_path)
                )

        try:
            yield (template_path, static_resc_paths, dynamic_resc_paths)
        finally:
            ctx_stk.pop_all().close()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.name}')"


class CustomTemplate(Template):
    def __init__(self, path: Path):
        if not path.exists():
            raise ValueError(f"custom template does not exist: '{path}'")
        self.path = path

    @contextmanager
    def load(self):
        yield (self.path, {}, {})

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.path}')"



import json
import subprocess
import sys
from pathlib import Path

# Template node.js script to read a file, process with liquidjs,
# and write the output to a file.
_NODE_TEMPLATE = """
const fs = require("fs");
var {{ Liquid }} = require({liquidjs_path});
var liquid = new Liquid({{
    jekyllinclude: true,
    dynamicPartials: false,
}});
metadata = {metadata};
data = fs.readFileSync({input_path}, "utf8");
data = liquid.parseAndRenderSync(data, metadata);
fs.writeFileSync({output_path}, data, "utf8");
"""

# Arguments to pass to `npm install` when installing liquidjs.
_NPM_INSTALL_ARGS = ("--no-save", "--no-fund", "--no-audit")


def run_liquidjs(input_path: Path, output_path: Path, install_dir: Path, **metadata):
    """Process a file using LiquidJS.

    Args:
        input_path: Path to input file.
        output_path: Path to output file.
        install_dir: Directory for installing LiquidJS (`node_modules` folder will be
            created here).
        metadata: Metadata to pass to LiquidJS.

    Raises:
        subprocess.CalledProcessError: If any subprocess fails.
    """
    liquidjs_path = (install_dir / "node_modules" / "liquidjs").absolute()
    if not liquidjs_path.exists():
        install_cmd = ["npm", "install", *_NPM_INSTALL_ARGS, "liquidjs"]
        print("+", *install_cmd, file=sys.stderr)
        subprocess.run(install_cmd, check=True, cwd=install_dir)

    node_script = _NODE_TEMPLATE.format(
        liquidjs_path=json.dumps(str(liquidjs_path)),
        metadata=json.dumps(metadata),
        input_path=json.dumps(str(input_path)),
        output_path=json.dumps(str(output_path)),
    )

    print("+ node <LIQUIDJS_TEMPLATE>", file=sys.stderr)
    subprocess.run(
        ["node", "-"],
        check=True,
        text=True,
        input=node_script,
        stdout=sys.stdout,
        stderr=sys.stderr,
    )



import subprocess
import sys
from pathlib import Path


def run_latexmk(
    input_path: Path, build_dir: Path, clean: bool = False, verbose: bool = False
) -> Path:
    """Compile a LaTeX file using `latexmk`.

    Args:
        input_path: Path to input file.
        build_dir: Path to build directory.
        clean: Whether to do a clean build by passing `-gg` to `latexmk`.
        verbose: Whether to enable verbose output.

    Returns:
        Path to output file.

    Raises:
        subprocess.CalledProcessError: If `latexmk` subprocess fails.
    """
    cmd = ["latexmk", "-pdf", "-interaction=nonstopmode", "-lualatex", "-f"]

    if clean:
        cmd.append("-gg")
    if not verbose:
        cmd.append("-silent")

    cmd.append(f"-output-directory={build_dir}")
    cmd.append(str(input_path))

    print("+", *cmd, file=sys.stderr)
    subprocess.run(
        cmd,
        check=True,
        stdout=sys.stdout if verbose else subprocess.DEVNULL,
        stderr=sys.stderr,
    )
    return build_dir / input_path.with_suffix(".pdf").name



import os
import shutil
import sys
from argparse import ArgumentParser
from contextlib import ExitStack, redirect_stderr, redirect_stdout
from pathlib import Path
from subprocess import CalledProcessError
from tempfile import NamedTemporaryFile
from typing import Annotated, Literal

import yaml
from corgy import Corgy, corgychecker, CorgyHelpFormatter, corgyparser, NotRequired
from corgy.types import KeyValuePairs

from ._latexmk import run_latexmk
from ._liquid import run_liquidjs
from ._pandoc import run_pandoc
from ._templates import BuiltinTemplate, Template
from ._version import __version__ as corgy_version


class ShinyMDC(Corgy, corgy_required_by_default=True):
    Flag = Literal[True]

    clean: NotRequired[Annotated[Flag, "do a clean build", ["-c", "--clean"]]]
    verbose: NotRequired[Annotated[Flag, "enable verbose output", ["-v", "--verbose"]]]
    quiet: NotRequired[Annotated[Flag, "suppress all output", ["-q", "--quiet"]]]

    input_path: NotRequired[
        Annotated[
            Path,
            "input markdown file (will read from standard input if not specified)",
            ["-i", "--input"],
        ]
    ]
    output_path: NotRequired[
        Annotated[
            Path,
            "output file (will write to standard output if not specified)",
            ["-o", "--output"],
        ]
    ]

    template: Annotated[
        Template,
        f"LaTeX template (path to custom template file, or name of a builtin "
        f"template: {'/'.join(BuiltinTemplate.BUILTIN_TEMPLATE_NAMES)})",
        ["-t", "--template"],
    ] = BuiltinTemplate("basic")

    meta: NotRequired[
        Annotated[
            KeyValuePairs,
            "extra metadata to update the document metadata with",
            ["-m", "--meta"],
        ]
    ]
    default_img_ext: Annotated[
        str, "default image extension", ["-x", "--default-img-ext"]
    ] = "pdf"
    natbib: Annotated[bool, "whether to use 'natbib' for bibliography"] = True

    build_dir: Annotated[
        Path,
        "directory for storing intermediate files and dependencies",
        ["-d", "--build-dir"],
    ] = Path(".shinymdc")
    tex_files_dir: NotRequired[
        Annotated[
            Path,
            "store main tex files in this directory instead of '<build-dir>/main'",
            ["-D", "--tex-files-dir"],
        ]
    ]
    main_tex_path: NotRequired[
        Annotated[
            Path, "write main tex to this file instead of '<tex-files-dir>/main.tex'",
        ]
    ]
    copy_statics: NotRequired[
        Annotated[Flag, "copy static resources to the tex files directory"]
    ]

    @corgychecker("input_path")
    @staticmethod
    def _check_is_readable_file(path: Path) -> Path:
        if not path.exists():
            raise ValueError(f"no such file: '{path}'")
        if not path.is_file():
            raise ValueError(f"not a file: '{path}'")
        if not os.access(path, os.R_OK):
            raise ValueError(f"file not readable: '{path}'")
        return path

    @corgychecker("output_path", "main_tex_path")
    @staticmethod
    def _check_is_writable_file(path: Path) -> Path:
        if path.exists():
            if not path.is_file():
                raise ValueError(f"not a file: '{path}'")
            if not os.access(path, os.W_OK):
                raise ValueError(f"file not writable: '{path}'")
        else:
            try:
                path.parent.mkdir(parents=True, exist_ok=True)
                path.touch()
            except OSError as e:
                raise ValueError(f"failed to create file: '{path}': {e}") from None
        return path

    @corgychecker("build_dir", "tex_files_dir")
    @staticmethod
    def _check_is_writable_dir(path: Path) -> Path:
        if path.exists():
            if not path.is_dir():
                raise ValueError(f"not a directory: '{path}'")
            if not os.access(path, os.W_OK):
                raise ValueError(f"directory not writable: '{path}'")
        else:
            try:
                path.mkdir(parents=True, exist_ok=True)
            except OSError as e:
                raise ValueError(f"failed to create directory: '{path}': {e}") from None
        return path

    @corgyparser("template", metavar="name|path")
    @staticmethod
    def _parse_template(arg: str) -> Template:
        return Template.make(arg)

    @corgyparser("input_path", metavar="file")
    @staticmethod
    def _parse_ifpath(arg: str) -> Path:
        return ShinyMDC._check_is_readable_file(Path(arg))

    @corgyparser("output_path", "main_tex_path", metavar="file")
    @staticmethod
    def _parse_ofpath(arg: str) -> Path:
        return ShinyMDC._check_is_writable_file(Path(arg))

    @corgyparser("build_dir", "tex_files_dir", metavar="dir")
    @staticmethod
    def _parse_odpath(arg: str) -> Path:
        return ShinyMDC._check_is_writable_dir(Path(arg))

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not hasattr(self, "tex_files_dir"):
            self.tex_files_dir = self.build_dir / "main"
            self.tex_files_dir.mkdir(exist_ok=True)
        if not hasattr(self, "main_tex_path"):
            self.main_tex_path = self.tex_files_dir / "main.tex"
            self.main_tex_path.touch()
        if not hasattr(self, "meta"):
            self.meta = KeyValuePairs("")

    def __call__(self) -> int:
        # Wrap the actual call with stdout/stderr redirections.
        # If 'output_path' is not specified, output has to be written to stdout.
        # So, redirect stdout to stderr to prevent subprocesses from writing
        # to stdout. The redirect is done unconditionaly to be consistent.
        # If 'quiet', redirect stderr to a temp file, so that the output can be
        # printed if there is an error.
        true_stdout = sys.stdout
        true_stderr = sys.stderr
        with ExitStack() as stack:
            if hasattr(self, "quiet"):
                # Redirect stderr and stdout to a temp file.
                temp_file = stack.enter_context(NamedTemporaryFile("r+"))
                stack.enter_context(redirect_stderr(temp_file))
                stack.enter_context(redirect_stdout(temp_file))
            else:
                # Redirect stdout to stderr.
                temp_file = None
                stack.enter_context(redirect_stdout(sys.stderr))

            try:
                self._call(true_stdout)
            except ValueError as e:
                print(f"ERROR: {e}", file=true_stderr)
                ret = 1
            except OSError as e:
                print(f"ERROR: {e.filename + ': ' or ''}{e.strerror}", file=true_stderr)
                ret = e.errno
            except CalledProcessError as e:
                ret = e.returncode
                print(f"ERROR: command failed (code {ret}):", *e.cmd, file=true_stderr)
            except KeyboardInterrupt:
                return -1
            else:
                ret = 0

            if ret != 0 and temp_file is not None:
                # Print stderr output from temp file.
                temp_file.seek(0)
                print(temp_file.read(), file=true_stderr)
        return ret

    def _call(self, stdout) -> None:
        ## 0. Clean up.
        if hasattr(self, "clean"):
            for _f in self.tex_files_dir.glob("*.tex"):
                if _f.is_file():
                    print(f"+ rm '{_f}'", file=sys.stderr)
                    _f.unlink()
            _rescs_dir = self.tex_files_dir / "resources"
            if _rescs_dir.is_dir():
                print(f"+ rm -rf '{_rescs_dir}'", file=sys.stderr)
                shutil.rmtree(_rescs_dir, ignore_errors=True)

        ## 1. Load input and parse metadata.

        # Read input from the input file or stdin.
        if hasattr(self, "input_path"):
            print(f"+ read '{self.input_path}'", file=sys.stderr)
            input_data = self.input_path.read_text()
        else:
            print("+ read stdin", file=sys.stderr)
            input_data = sys.stdin.read()

        # Extract header from input data.
        input_lines_iter = iter(input_data.splitlines())
        if next(input_lines_iter) != "---":
            raise ValueError
        header = ""
        try:
            while (line := next(input_lines_iter)) != "---":
                header += line + "\n"
        except StopIteration:
            raise ValueError from None

        # Rest of the input data is the main body.
        body_md = "\n".join(input_lines_iter).strip()

        # Parse metadata from header.
        metadata = yaml.safe_load(header)
        if metadata is None:
            metadata = {}
        try:
            bib_path = Path(metadata["bibliography"])
        except KeyError:
            bib_path = None

        # Link authors to institutes using institute IDs. This is for templates
        # which don't use superscripts for author affiliations.
        # Institutes are specified as a list of dicts with keys 'id' and 'name'.
        # Author affiliations are specified as a list of institute IDs.
        institutes = metadata.get("institute", [])
        inst_id_to_name = {_inst["id"]: _inst["name"] for _inst in institutes}
        authors = metadata.get("author", [])
        for author in authors:
            author["institute"] = []
            for inst_id in author.get("affiliation", []):
                try:
                    author["institute"].append(inst_id_to_name[inst_id])
                except KeyError:
                    raise ValueError(
                        f"unknown institute id '{inst_id}' in author '{author['name']}'"
                    ) from None
        metadata["authors"] = authors

        # Set 'skipequal' to True if no authors have equal contribution.
        if metadata.get("skipequal", False) is False:
            metadata["skipequal"] = (
                sum(int(author.get("equalcontrib", False)) for author in authors) < 2
            )

        # Get the longest author name, email, and institute name.
        # This can be used by templates to align the author list.
        metadata["longest_author_name"] = (
            max((_a.get("name", "") for _a in authors), key=len) if authors else 0
        )
        metadata["longest_email"] = (
            max((_a.get("email", "") for _a in authors), key=len) if authors else 0
        )
        metadata["longest_inst_name"] = (
            max((_i.get("name", "") for _i in institutes), key=len) if institutes else 0
        )

        ############################################################

        ## 2. Prepare input for first stage processing.

        # Create combined input for first stage processing.
        # This input combines the body, abstract, includes, and appendices.
        # The pieces are separated by a non-printing unicode character, so they can
        # be split apart later.
        primary_sep = "\n\n\ufdd0\n\n"
        secondary_sep = "\n\n\ufdd1\n\n"

        combined_md = body_md + primary_sep
        combined_md += metadata.get("abstract", "") + primary_sep

        def _read_files_and_add_to_combined(pstrs: list[str]):
            nonlocal combined_md
            for _i, _pstr in enumerate(pstrs):
                # Separate each piece by a non-printing character.
                if _i != 0:
                    combined_md += secondary_sep
                print(f"+ read '{_pstr}'", file=sys.stderr)
                combined_md += Path(_pstr).read_text()
            combined_md += primary_sep

        for _kind in ["sections", "includes", "appendices"]:
            _read_files_and_add_to_combined(metadata.get(_kind, []))

        # Save combined markdown to a file.
        combined_md_path = Path(self.build_dir / "combined.md")
        print(f"+ write '{combined_md_path}'", file=sys.stderr)
        combined_md_path.write_text(combined_md)

        ############################################################

        ## 3. Generate raw latex for input.

        # Process combined markdown with liquidjs.
        combined_liquified_md_path = self.build_dir / "combined.liquified.md"
        run_liquidjs(
            combined_md_path, combined_liquified_md_path, self.build_dir, **self.meta
        )

        # Process liquidjs output with pandoc.
        with NamedTemporaryFile(suffix=".tex") as stub_template_file:
            stub_template_path = Path(stub_template_file.name)
            stub_template_path.write_text("$body$")
            combined_tex = run_pandoc(
                combined_liquified_md_path,
                stub_template_path,
                bib_path,
                None,  # metadata_path
                self.default_img_ext,
                self.natbib,
                hasattr(self, "verbose"),
                **self.meta,
            )

        # Split compiled latex into constituent parts.
        parts = combined_tex.split(primary_sep.strip())
        body_tex = parts[0].strip()
        abstract_tex = parts[1].strip()
        sections_tex = parts[2].split(secondary_sep.strip())
        includes_tex = parts[3].split(secondary_sep.strip())
        appendices_tex = parts[4].split(secondary_sep.strip())

        ############################################################

        ## 4. Link raw latex parts into metadata dictionary for final processing.

        # The different parts are written to files, and 'include' statements linking
        # to them are added to the metadata dictionary. This is to prevent pandoc
        # from reprocessing them.

        _input_template = r"\input{%s}"

        # Save body.
        if body_tex:
            body_tex_path = self.tex_files_dir / "body.tex"
            print(f"+ write '{body_tex_path}'", file=sys.stderr)
            body_tex_path.write_text(body_tex)
            metadata["body"] = _input_template % body_tex_path.with_suffix("")

        # Save abstract.
        if abstract_tex:
            abstract_tex_path = self.tex_files_dir / "abstract.tex"
            print(f"+ write '{abstract_tex_path}'", file=sys.stderr)
            abstract_tex_path.write_text(abstract_tex)
            metadata["abstract"] = _input_template % abstract_tex_path.with_suffix("")

        def _write_tex_and_get_pstrs(texs, metadata_key, otype):
            new_pstrs = []
            for _tex, _old_pstr in zip(texs, metadata.get(metadata_key, [])):
                _tex = _tex.strip()
                if not _tex:
                    continue
                old_fname = Path(_old_pstr).stem
                new_tex_path = self.tex_files_dir / f"{otype}_{old_fname}.tex"
                print(f"+ write '{new_tex_path}'", file=sys.stderr)
                new_tex_path.write_text(_tex)
                new_pstrs.append(str(new_tex_path.with_suffix("")))
            return new_pstrs

        section_tex_pstrs = _write_tex_and_get_pstrs(
            sections_tex, "sections", "section"
        )
        include_tex_pstrs = _write_tex_and_get_pstrs(
            includes_tex, "includes", "include"
        )
        appendix_tex_pstrs = _write_tex_and_get_pstrs(
            appendices_tex, "appendices", "appendix"
        )

        for _kind, _pstrs in zip(
            ["sections", "includes", "appendices"],
            [section_tex_pstrs, include_tex_pstrs, appendix_tex_pstrs],
        ):
            metadata[_kind] = [_input_template % _pstr for _pstr in _pstrs]

        # Write metadata to file.
        metadata_path = self.build_dir / "metadata.yaml"
        print(f"+ write '{metadata_path}'", file=sys.stderr)
        metadata_path.write_text(yaml.dump(metadata))

        ############################################################

        ## 5. Combine raw latex parts with user template.

        # Now, all the necessary parts are linked through the metadata.
        # So we can pass a dummy empty file to pandoc with the user template.
        rescs_dir = self.tex_files_dir / "resources"
        rescs_dir.mkdir(exist_ok=True)
        with self.template.load() as (
            template_path,
            static_resc_paths,
            dynamic_resc_paths,
        ):
            # Add static resources to metadata.
            for resc_name, resc_path in static_resc_paths.items():
                if hasattr(self, "copy_statics"):
                    # Copy the resource to the tex files dir if it is not already there.
                    copy_resc_path = rescs_dir / f"{resc_name}{resc_path.suffix}"
                    if not copy_resc_path.exists():
                        print(f"+ cp '{resc_path}' '{copy_resc_path}'", file=sys.stderr)
                        shutil.copy(resc_path, copy_resc_path)
                    resc_path = copy_resc_path
                metadata[resc_name] = str(resc_path.with_suffix(""))

            # Process dynamic resources with pandoc.
            for resc_name, resc_path in dynamic_resc_paths.items():
                processed_resc_tex = run_pandoc(
                    Path(os.devnull),
                    resc_path,
                    bib_path,
                    metadata_path,
                    self.default_img_ext,
                    self.natbib,
                    hasattr(self, "verbose"),
                    **self.meta,
                )
                processed_resc_path = rescs_dir / f"{resc_name}{resc_path.suffix}"
                print(f"+ write '{processed_resc_path}'", file=sys.stderr)
                processed_resc_path.write_text(processed_resc_tex)
                metadata[resc_name] = str(processed_resc_path.with_suffix(""))

            # Write updated metadata to file.
            print(f"+ write '{metadata_path}'", file=sys.stderr)
            metadata_path.write_text(yaml.dump(metadata))

            # Process main tex with pandoc.
            main_tex = run_pandoc(
                Path(os.devnull),
                template_path,
                bib_path,
                metadata_path,
                self.default_img_ext,
                self.natbib,
                hasattr(self, "verbose"),
                **self.meta,
            )

        # Write main tex to file.
        print(f"+ write '{self.main_tex_path}'", file=sys.stderr)
        self.main_tex_path.write_text(main_tex)

        ############################################################

        ## 6. Generate pdf using latexmk.

        # Compile main tex with latexmk.
        latexmk_build_dir = self.build_dir / "latexmk"
        print(f"+ mkdir -p '{latexmk_build_dir}'", file=sys.stderr)
        latexmk_build_dir.mkdir(exist_ok=True)
        result_path = run_latexmk(
            self.main_tex_path,
            latexmk_build_dir,
            hasattr(self, "clean"),
            hasattr(self, "verbose"),
        )

        # Copy result to output file or write it to stdout.
        if hasattr(self, "output_path"):
            print(f"+ cp '{result_path}' '{self.output_path}'", file=sys.stderr)
            shutil.copy(result_path, self.output_path)
        else:
            print(f"+ cat '{result_path}'", file=sys.stderr)
            stdout.buffer.write(result_path.read_bytes())


def main():
    parser = ArgumentParser(formatter_class=CorgyHelpFormatter)
    parser.add_argument("-V", "--version", action="version", version=corgy_version)
    mdc = ShinyMDC.parse_from_cmdline(parser)
    return mdc()
